#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#include <unistd.h>
#include <limits.h>

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t condition = PTHREAD_COND_INITIALIZER;
pthread_rwlock_t lock;

#define NUMBERS 100
int numbers[NUMBERS];
int averageRun = 0;
int sumRun = 0;
int maxRun = 0;

void * makerandom(void* spare)
{
	while(1)
	{
		pthread_mutex_lock(&mutex);
		pthread_rwlock_wrlock(&lock);
		printf("Generating numbers...\n");
		for(unsigned int i=0 ; i<NUMBERS ; i++)
		{
			numbers[i] = random();
		}
		//Unlock the array when we have finished with it.
		pthread_rwlock_unlock(&lock);
		//Set a flag so that the other functions run if they miss the condition variable
		maxRun = averageRun = sumRun = 1;
		//Wake everyone else up...
		pthread_cond_broadcast(&condition);
		//...and let them run
		pthread_mutex_unlock(&mutex);
		sleep(1);
	}
}

void * sum(void * spare)
{
	while(1)
	{
		pthread_mutex_lock(&mutex);
		//Only wait on the condition variable if there is nothing else to do
		if(sumRun == 0)
		{
			pthread_cond_wait(&condition, &mutex);
		}
		//Protect the array while we look at it
		pthread_rwlock_rdlock(&lock);
		if(sumRun == 1)
		{
			//Unset the flag before releasing the mutex so we don't miss it being set again
			sumRun = 0;
			pthread_mutex_unlock(&mutex);

			long long total = 0;
			for(unsigned int i=0 ; i<NUMBERS ; i++)
			{
				total += numbers[i];
			}
			printf("Total = %lld\n",total); 
		}
		else
		{
			//Release the mutex if we didn't do anything with the array
			pthread_mutex_unlock(&mutex);
		}
		sleep(1);
		//Unlock the array
		pthread_rwlock_unlock(&lock);
	}
}

void * average(void * spare)
{
	while(1)
	{
		pthread_mutex_lock(&mutex);
		if(averageRun == 0)
		{
			pthread_cond_wait(&condition, &mutex);
		}
		pthread_rwlock_rdlock(&lock);
		if(averageRun == 1)
		{
			averageRun = 0;
			pthread_mutex_unlock(&mutex);
			long long total = 0;
			for(unsigned int i=0 ; i<NUMBERS ; i++)
			{
				total += numbers[i];
			}
			printf("Average = %lld\n",total / NUMBERS); 
		}
		else
		{
			pthread_mutex_unlock(&mutex);
		}
		pthread_rwlock_unlock(&lock);
	}
}
void * max(void * spare)
{
	while(1)
	{
		pthread_mutex_lock(&mutex);
		if(maxRun == 0)
		{
			pthread_cond_wait(&condition, &mutex);
		}
		pthread_rwlock_rdlock(&lock);
		if(maxRun == 1)
		{
			maxRun = 0;
			pthread_mutex_unlock(&mutex);
			int max = -INT_MAX;
			for(unsigned int i=0 ; i<NUMBERS ; i++)
			{
				if(numbers[i] > max)
				{
					max = numbers[i];
				}
			}
			printf("Max = %d\n",max); 
		}
		else
		{
			pthread_mutex_unlock(&mutex);
		}
		pthread_rwlock_unlock(&lock);
	}
}
int main(void)
{
	pthread_t thread1, thread2;
	//Dynamically initialise the rwlock with default attributes
	pthread_rwlock_init(&lock, NULL);
	pthread_create(&thread1, NULL, sum, (void*)1);
	pthread_create(&thread1, NULL, average, (void*)2);
	pthread_create(&thread1, NULL, max, (void*)3);
	pthread_create(&thread2, NULL, makerandom, NULL);
	pthread_join(thread1, NULL);
	pthread_join(thread2, NULL);
	return 0;
}
